/**
 * CDataPair<br>
 * A class to simplify storage, manipulation and retrieval of name-value pairs. One
 * restriction is that names or values cannot be empty strings.
 * <br>
 * You are free to use this code and to make modifications provided
 * this notice is retained.
 * <p>
 * If you found this useful, please add a note of acknowledgement to my guestbook,
 * or send me a mail at <a href="mailto:tanmaykm@hotmail.com">tanmaykm@hotmail.com</a>. 
 * If you would like to report a bug or suggest some improvements, 
 * you are most welcome. I will be happy to help you use this piece of code.
 * <p>
 *
 * @author 	    Tanmay K. Mohapatra
 * @version     1.01, 18th Dec, 1999
 *
 * Modification log:
 * 1.00  Tanmay  27th Jun 1999  Original Version
 * 1.01  Tanmay  18th Dec 1999  Fixed bug in findEntry(int) function
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "DataPair.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDataPair::CDataPair()
{
    pList = NULL;
    iNumElements = 0;
}

CDataPair::~CDataPair()
{
    removeAll();
}

int CDataPair::removeAll(void)
{
    struct stDataPair *pThis, *pPrev;
    for(pThis=pList; pThis; )
    {
        if(pThis->pName)    free(pThis->pName);
        if(pThis->pValue)   free(pThis->pValue);
        pPrev = pThis;
        pThis=pThis->pNext;
        free(pPrev);
    }
    pList = NULL;
    iNumElements = 0;
    return 0;
}

int CDataPair::parseData(const char *sData)
{
    char    *pName, *pValue, *pDupStr;

    pDupStr = strdup(sData);
    
    for(pName=strtok(pDupStr, "="); pName; pName=strtok(NULL, "="))
    {
        if(pName[strlen(pName)+1] != '&')
        {
            if (!(pValue=strtok(NULL, "&"))) break;
        }
        else
        {
            pValue = "+";
        }
        setNameValuePair(pName, pValue);
    }
    free(pDupStr);
    return 0;
}

char * CDataPair::getAsString(void)
{
    char *sRet;
    struct stDataPair   *pThis;
    int                 iStrLen;

    for(pThis=pList, iStrLen=0; pThis; pThis=pThis->pNext)
    {
        iStrLen += strlen(pThis->pName) + strlen(pThis->pValue) + 2;
    }

    if (iStrLen <= 0)
    {
        return 0;
    }

    sRet = (char *)malloc(iStrLen+1);
    if (!sRet)  return 0;

    for(pThis=pList, iStrLen=0; pThis; pThis=pThis->pNext)
    {
        iStrLen += sprintf(sRet+iStrLen, "%s=%s&", pThis->pName, pThis->pValue);
    }
    sRet[iStrLen-1] = 0;

    return sRet;
}

char * CDataPair::getValueFor(const char *sName)
{
    struct stDataPair *pThis;

    pThis = findEntry(sName);
    if (pThis)
    {
        return pThis->pValue;
    }
    return NULL;
}

int CDataPair::getLengthFor(const char *sName)
{
    struct stDataPair *pThis;

    pThis = findEntry(sName);
    if (pThis)
    {
        return strlen(pThis->pValue);
    }
    return -1;
}

int CDataPair::isPresent(const char *sName)
{
    return findEntry(sName) ? 1 : 0;
}

int CDataPair::getValueFor(const char *sName, char *sValue, int iLength)
{
    struct stDataPair *pThis;

    pThis = findEntry(sName);
    if (pThis)
    {
        strncpy(sValue, pThis->pValue, iLength-1);
        sValue[iLength] = 0;
        return 0;
    }
    return -1;
}

int CDataPair::getValueFor(const char *sName, char **sValue, int *piLength)
{
    struct stDataPair *pThis;

    pThis = findEntry(sName);
    if (pThis)
    {
        if (*piLength > (int)strlen(pThis->pValue))
        {
            strcpy(*sValue, pThis->pValue);
            return 0;
        }
        *sValue = (char *)realloc(*sValue, *piLength=strlen(pThis->pValue)+1);
        strcpy(*sValue, pThis->pValue);
        return 0;
    }
    return -1;
}

int CDataPair::getNameValueAt(char *sName, char *sValue, int iLength, int iIndex)
{
    struct stDataPair *pThis;

    pThis = findEntry(iIndex);
    if (pThis)
    {
        strcpy(sName, pThis->pName);
        strncpy(sValue, pThis->pValue, iLength-1);
        sValue[iLength] = 0;
        return 0;
    }
    return -1;
}

int CDataPair::getNameValueAt(char *sName, char **sValue, int *piLength, int iIndex)
{
    struct stDataPair *pThis;

    pThis = findEntry(iIndex);
    if (pThis)
    {
        strcpy(sName, pThis->pName);
        if (*piLength > (int)strlen(pThis->pValue))
        {
            strcpy(*sValue, pThis->pValue);
            return 0;
        }
        *sValue = (char *)realloc(*sValue, *piLength=strlen(pThis->pValue)+1);
        strcpy(*sValue, pThis->pValue);
        return 0;
    }
    return -1;
}

int CDataPair::setNameValuePair(const char *sName, const char *sValue)
{
    struct stDataPair *pThis;

    pThis = findEntry(sName);
    if (pThis)
    {
        if(pThis->pValue) free(pThis->pValue);
        pThis->pValue = strdup(sValue);
    }
    else
    {
        pThis = (struct stDataPair *)malloc(sizeof(*pThis));
        if (!pThis) return -1;
        pThis->pName = strdup(sName);
        pThis->pValue = strdup(sValue);
        pThis->pNext = pList;
        pList = pThis;
        iNumElements++;
    }
    return 0;
}

int CDataPair::escapeData(void)
{
    struct stDataPair   *pThis;
    char                *pNewStr;
    int                 iStrLen, iCount, iNewCount;

    for(pThis=pList; pThis; pThis=pThis->pNext)
    {
        iStrLen = strlen(pThis->pValue);
        pNewStr = (char *)malloc(3*iStrLen*sizeof(char)+1);
        if (!pNewStr) return -1;
        for (iCount=iNewCount=0; iCount<iStrLen; iCount++)
        {
            if( ((pThis->pValue[iCount] >= '0') && (pThis->pValue[iCount] <= '9')) || 
                ((pThis->pValue[iCount] >= 'a') && (pThis->pValue[iCount] <= 'z')) ||
                ((pThis->pValue[iCount] >= 'A') && (pThis->pValue[iCount] <= 'Z')) ||
                (pThis->pValue[iCount] == '-')  || (pThis->pValue[iCount] == '_')  ||
                (pThis->pValue[iCount] == '.')  || (pThis->pValue[iCount] == '*')  )
            {
                pNewStr[iNewCount] = pThis->pValue[iCount];
                iNewCount++;
            }
            else if(pThis->pValue[iCount] == ' ')
            {
                pNewStr[iNewCount] = '+';
                iNewCount++;
            }
            else
            {
                pNewStr[iNewCount] = '%';
                sprintf((pNewStr + iNewCount + 1), "%02x", ((int) pThis->pValue[iCount]));
                iNewCount+=3;
            }
        }
        pNewStr[iNewCount] = '\0';
        free(pThis->pValue);
        pThis->pValue = pNewStr;
    }
    return 0;
}

int CDataPair::unescapeData(void)
{
    struct stDataPair   *pThis;
    int                 iStrLen, iCount, iNewCount;

    for(pThis=pList; pThis; pThis=pThis->pNext)
    {
        iStrLen = strlen(pThis->pValue);
        for (iCount=iNewCount=0; iCount<iStrLen; iCount++)
        {
            if( ((pThis->pValue[iCount] >= '0') && (pThis->pValue[iCount] <= '9')) || 
                ((pThis->pValue[iCount] >= 'a') && (pThis->pValue[iCount] <= 'z')) ||
                ((pThis->pValue[iCount] >= 'A') && (pThis->pValue[iCount] <= 'Z')) ||
                (pThis->pValue[iCount] == '-')  || (pThis->pValue[iCount] == '_')  ||
                (pThis->pValue[iCount] == '.')  || (pThis->pValue[iCount] == '*')  )
            {
                pThis->pValue[iNewCount] = pThis->pValue[iCount];
                iNewCount++;
            }
            else if(pThis->pValue[iCount] == '+')
            {
                pThis->pValue[iNewCount] = ' ';
                iNewCount++;
            }
            else if(pThis->pValue[iCount] == '%')
            {
                if (pThis->pValue[iCount+1] == '%')
                {
                    pThis->pValue[iNewCount] = pThis->pValue[iCount];
                    iNewCount++;
                    iCount++;
                    pThis->pValue[iNewCount] = pThis->pValue[iCount];
                    iNewCount++;
                }
                else
                {
                    int iVal;
                    char cVal;
                    cVal = pThis->pValue[iCount+1];
                    if((cVal >= '0') && (cVal <= '9'))          iVal = 16*(cVal-'0');
                    else if ((cVal >= 'a') && (cVal <= 'z'))    iVal = 16*(cVal-'a'+10);
                    else                                        iVal = 16*(cVal-'A'+10);
                    cVal = pThis->pValue[iCount+2];
                    if((cVal >= '0') && (cVal <= '9'))          iVal += (cVal-'0');
                    else if ((cVal >= 'a') && (cVal <= 'z'))    iVal += (cVal-'a'+10);
                    else                                        iVal += (cVal-'A'+10);
                    pThis->pValue[iNewCount] = iVal;
                    iNewCount++;
                    iCount += 2;
                }
            }
            else
            {
                pThis->pValue[iNewCount] = pThis->pValue[iCount];
                iNewCount++;
            }
        }
        pThis->pValue[iNewCount] = '\0';
    }
    return 0;
}

int CDataPair::getNumberOfPairs(void)
{
    return iNumElements;
}

struct stDataPair * CDataPair::findEntry(const char *sName)
{
    struct stDataPair *pThis;
    for(pThis=pList; pThis; pThis=pThis->pNext)
    {
        if (!strcmp(pThis->pName, sName)) return pThis;
    }
    return NULL;
}

struct stDataPair * CDataPair::findEntry(int iIndex)
{
    int iThisIndex;
    struct stDataPair *pThis = NULL;
    for(pThis=pList, iThisIndex=0; pThis && iThisIndex<iIndex; pThis=pThis->pNext, iThisIndex++);
   
    return (iThisIndex == iIndex) ? pThis : NULL;
}
